##
# File:    ScanRepoUtil.py
# Author:  J. Westbrook
# Date:    30-Apr-2018
# Version: 0.001
#
# Updates:
#  9-May-2018 jdw implement incremental update of scanned data
# 18-May-2018 jdw move IO to IoUtils.py
# 23-May-2018 jdw remove proxy to ContentTypeUtil(), go directly to RepoPathUtil() for paths
#  5-Jun-2018 jdw update prototypes for IoUtil() methods
# 16-Jun-2018 jdw update data type prototype.
# 18-Jun-2018 jdw move mocking to configuration level
# 28-Jun-2018 jdw remove IoUtil() and add working path constructor argument
# 14-Feb-2020 jdw migrate to rcsb.utils.repository
##
"""
Tools for for scanning repositories and collecting coverage and type data information.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import collections
import datetime
import logging
import os
import time

from mmcif.api.DataCategory import DataCategory
from rcsb.utils.repository.RepositoryProvider import RepositoryProvider
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)

ScanValue = collections.namedtuple("ScanValue", "containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec")
ScanSummary = collections.namedtuple("ScanSummary", "containerId, fromPath, scanDate, scanCategoryDict")
ChemSummary = collections.namedtuple("ChemSummary", "entryId, releaseDate, ccTup")


class ScanRepoUtil(object):
    """Tools for for scanning repositories and collecting coverage and type data information."""

    def __init__(self, cfgOb, attributeDataTypeD=None, numProc=4, chunkSize=15, fileLimit=None, maxStepLength=2000, workPath=None):
        """
        Args:
            cfgOb (object): Configuration object (rcsb.utils.config.ConfigUtil)

            attributeDataTypeD
            dictPath (str): Path to supporting data dictionary

            numProc (int, optional): Number of parallel worker processes used.
            chunkSize (int, optional): Size of files processed in a single multi-proc process
            fileLimit (int, optional): maximum file scanned or None for no limit
            mockTopPath (str, optional): Path to directory containing mock repositories or None
            maxStepLength (int, optional): maximum number of multi-proc runs to perform
        """
        #
        self.__attributeDataTypeD = attributeDataTypeD if attributeDataTypeD else {}
        # Limit the load length of each file type for testing  -  Set to None to remove -
        self.__fileLimit = fileLimit
        self.__maxStepLength = maxStepLength
        #
        # Controls for multiprocessing execution -
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        #
        self.__cfgOb = cfgOb
        #
        self.__mpFormat = "[%(levelname)s] %(asctime)s %(processName)s-%(module)s.%(funcName)s: %(message)s"

        self.__workPath = workPath
        self.__mU = MarshalUtil(workPath=self.__workPath)
        self.__rpP = RepositoryProvider(self.__cfgOb, numProc=self.__numProc, fileLimit=self.__fileLimit, cachePath=self.__workPath)

    def scanContentType(
        self,
        contentType,
        mergeContentTypes=None,
        scanType="full",
        inputPathList=None,
        scanDataFilePath=None,
        failedFilePath=None,
        saveInputFileListPath=None,
    ):
        """Driver method for repository scan operation

        Args:
            contentType (str):  one of 'bird','bird_family','bird_chem_comp', chem_comp','pdbx', 'pdbx_obsolete'
            scanType (str, optional): 'full' [or 'incr' to be supported]
            inputPathList (list, optional):  list of input file paths to scan
            scanDataFilePath (str, optional): file path for serialized scan data (Pickle format)
            failedFilePath (str, optional): file path for list of files that fail scanning operation
            saveInputFileListPath str, optional): Path to store file path list that is scanned

        Returns:
            bool: True for success or False otherwise

        """
        ok = False
        retLists = None
        try:
            startTime = self.__begin(message="scanning operation")
            #
            locatorObjList = self.__rpP.getLocatorObjList(contentType=contentType, inputPathList=inputPathList, mergeContentTypes=mergeContentTypes)
            #
            if saveInputFileListPath:
                self.__mU.doExport(saveInputFileListPath, self.__rpP.getLocatorPaths(locatorObjList), fmt="list")
                logger.debug("Saving %d paths in %s", len(locatorObjList), saveInputFileListPath)
            #
            optD = {}
            optD["contentType"] = contentType
            optD["logSize"] = True
            optD["scanType"] = scanType
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            #
            numProc = self.__numProc
            chunkSize = self.__chunkSize if locatorObjList and self.__chunkSize < len(locatorObjList) else 0
            #
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            numPaths = len(locatorObjList)
            logger.debug("Processing %d total paths", numPaths)
            numProc = min(numProc, numPaths)
            maxStepLength = self.__maxStepLength
            if numPaths > maxStepLength:
                numLists = int(numPaths / maxStepLength)
                subLists = [locatorObjList[i::numLists] for i in range(numLists)]
            else:
                subLists = [locatorObjList]
            #
            if subLists:
                logger.debug("Starting with numProc %d outer subtask count %d subtask length ~ %d", numProc, len(subLists), len(subLists[0]))
            #
            numResults = 2 if contentType.startswith("pdbx") else 1
            failList = []
            retLists = [[] for ii in range(numResults)]
            diagList = []
            for ii, subList in enumerate(subLists):
                logger.info("Running outer subtask %d of %d length %d", ii + 1, len(subLists), len(subList))
                #
                mpu = MultiProcUtil(verbose=True)
                mpu.setOptions(optionsD=optD)
                mpu.set(workerObj=self, workerMethod="scanWorker")
                ok, failListT, retListsT, diagListT = mpu.runMulti(dataList=subList, numProc=numProc, numResults=numResults, chunkSize=chunkSize)
                failList.extend(failListT)
                # retLists is a list of lists -
                logger.debug("status %r fail len %r ret len %r", ok, len(failListT), len(retListsT))
                for jj in range(numResults):
                    retLists[jj].extend(retListsT[jj])
                diagList.extend(diagListT)
            logger.debug("Scan failed path list %r", failList)
            logger.debug("Scan path list success length %d load list failed length %d", len(locatorObjList), len(failList))
            logger.debug("Returned metadata length %r", len(retLists[0]))
            #
            if failedFilePath and failList:
                wOk = self.__mU.doExport(failedFilePath, self.__rpP.getLocatorPaths(failList), fmt="list")
                logger.debug("Writing scan failure path list to %s status %r", failedFilePath, wOk)
            #
            if scanType == "incr":
                scanDataD = self.__mU.doImport(scanDataFilePath, fmt="pickle", default=None)
                logger.debug("Imported scan data with keys %r", list(scanDataD.keys()))
            else:
                scanDataD = {}
            #
            if scanDataFilePath and retLists[0]:
                for ssTup in retLists[0]:
                    cId = ssTup.containerId
                    if scanType == "full" and cId in scanDataD:
                        logger.error("Duplicate container id %s in %r and %r", cId, ssTup.fromPath, scanDataD[cId].fromPath)
                    #
                    scanDataD[cId] = ssTup

                ok = self.__mU.doExport(scanDataFilePath, scanDataD, fmt="pickle")
                tscanDataD = self.__mU.doImport(scanDataFilePath, fmt="pickle")
                ok = tscanDataD == scanDataD

            if scanDataFilePath and contentType.startswith("pdbx") and len(retLists) > 1 and retLists[1]:
                pth, _ = os.path.split(scanDataFilePath)
                fp = os.path.join(pth, "%s-chem-comp-release-data.json" % contentType)
                ok1 = self.__mU.doExport(fp, retLists[1], fmt="json")
                ok = ok and ok1

            self.__end(startTime, "scanning operation with status " + str(ok))

            #
            return ok
        except Exception as e:
            logger.exception("Failing %r with %s", contentType, str(e))

        return False

    def evalScan(self, scanDataFilePath, evalJsonFilePath, evalType="data_type"):

        scanDataD = self.__mU.doImport(scanDataFilePath, fmt="pickle")
        if evalType in ["data_type"]:
            rD = self.__evalScanDataType(scanDataD)
        elif evalType in ["data_coverage"]:
            rD, _ = self.__evalScanDataCoverage(scanDataD)
        else:
            logger.debug("Unknown evalType %r", evalType)
        ok = self.__mU.doExport(evalJsonFilePath, rD, fmt="json")

        return ok

    def evalScanItem(self, scanDataFilePath, evalFilePath):
        scanDataD = self.__mU.doImport(scanDataFilePath, fmt="pickle")
        _, cL = self.__evalScanDataCoverage(scanDataD)
        ok = self.__mU.doExport(evalFilePath, cL, fmt="list")
        return ok

    def __evalScanDataType(self, scanDataD):
        """
        ScanValue = collections.namedtuple('ScanValue', 'containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec')
        ScanSummary = collections.namedtuple('ScanSummary', 'containerId, fromPath, scanDate, scanCategoryDict')

        """
        # for populated sD[category] -> d[atName]->{minWidth: , maxWidth:, minPrec:, maxPrec: , count}
        sD = {}
        for cId in scanDataD:
            ssTup = scanDataD[cId]
            dD = ssTup.scanCategoryDict
            for catName in dD:
                if catName not in sD:
                    sD[catName] = {}
                for svTup in dD[catName]:
                    if svTup.atName not in sD[catName]:
                        sD[catName][svTup.atName] = {"minWidth": svTup.minWidth, "maxWidth": svTup.maxWidth, "minPrec": svTup.minPrec, "maxPrec": svTup.maxPrec, "count": 1}
                        continue
                    sD[catName][svTup.atName]["minWidth"] = min(sD[catName][svTup.atName]["minWidth"], svTup.minWidth)
                    sD[catName][svTup.atName]["maxWidth"] = max(sD[catName][svTup.atName]["maxWidth"], svTup.maxWidth)
                    sD[catName][svTup.atName]["minPrec"] = min(sD[catName][svTup.atName]["minPrec"], svTup.minPrec)
                    sD[catName][svTup.atName]["maxPrec"] = max(sD[catName][svTup.atName]["maxPrec"], svTup.maxPrec)
                    sD[catName][svTup.atName]["count"] += 1
        return sD

    def __evalScanDataCoverage(self, scanDataD):
        """
        ScanValue = collections.namedtuple('ScanValue', 'containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec')
        ScanSummary = collections.namedtuple('ScanSummary', 'containerId, fromPath, scanDate, scanCategoryDict')

        """

        # for populated sD[category] -> d[atName]->{count: #, instances: [id,id,id]}
        sD = {}
        for cId in scanDataD:
            ssTup = scanDataD[cId]
            dD = ssTup.scanCategoryDict
            for catName in dD:
                if catName not in sD:
                    sD[catName] = {}
                for svTup in dD[catName]:
                    if svTup.atName not in sD[catName]:
                        sD[catName][svTup.atName] = {"count": 0, "instances": []}
                    sD[catName][svTup.atName]["instances"].append(svTup.containerId)
                    sD[catName][svTup.atName]["count"] += 1
        cL = []
        for catName, aD in sD.items():
            for atName, tD in aD.items():
                cL.append("%s\t%s" % ("_" + catName + "." + atName, tD["count"]))
        return sD, cL

    def scanWorker(self, dataList, procName, optionsD, workingDir):
        """Multi-proc worker method for scanning repository data files-"""
        try:
            _ = workingDir
            startTime = self.__begin(message=procName)
            # Recover common options

            scanType = optionsD["scanType"]
            contentType = optionsD["contentType"]
            #
            successList = []
            retList = []
            chmList = []

            containerList = self.__getContainerList(dataList)
            for container in containerList:
                ret = self.__scanContainer(container)
                successList.append(ret.fromPath)
                retList.append(ret)
                if contentType.startswith("pdbx"):
                    chm = self.__scanChemistry(container)
                    chmList.append(chm)
            #

            logger.debug("%s scanType %s contentType %s pathlist length %d containerList length %d", procName, scanType, contentType, len(dataList), len(containerList))

            ok = len(successList) == len(dataList)
            #
            self.__end(startTime, procName + " with status " + str(ok))
            if contentType.startswith("pdbx"):
                return successList, retList, chmList, []
            else:
                return successList, retList, []

        except Exception as e:
            logger.error("Failing with dataList %r", dataList)
            logger.exception("Failing with %s", str(e))

        return [], [], []

    def __getContainerList(self, locatorObjList):
        """ """
        utcnow = datetime.datetime.utcnow()
        ts = utcnow.strftime("%Y-%m-%d:%H:%M:%S")
        cL = []
        myContainerList = self.__rpP.getContainerList(locatorObjList)
        for loc in locatorObjList:
            myContainerList = self.__rpP.getContainerList([loc])
            lPathL = self.__rpP.getLocatorPaths([loc])
            for cA in myContainerList:
                dc = DataCategory("rcsb_load_status", ["name", "load_date", "locator"], [[cA.getName(), ts, lPathL[0]]])
                logger.debug("data category %r", dc)
                cA.append(dc)
                cL.append(cA)
        return cL

    def __scanChemistry(self, container):
        """Get release date and the unique list of chemical components in the input container

        Args:
            container (obj): instance of an mmCIF data container

            loop_
            _pdbx_audit_revision_history.ordinal
            _pdbx_audit_revision_history.data_content_type
            _pdbx_audit_revision_history.major_revision
            _pdbx_audit_revision_history.minor_revision
            _pdbx_audit_revision_history.revision_date
            1 'Structure model' 1 0 1996-12-23
            2 'Structure model' 1 1 2008-03-24
            3 'Structure model' 1 2 2011-07-13
        """
        releaseDate = entryId = None
        ccS = set()
        if container.exists("entry"):
            tObj = container.getObj("entry")
            entryId = tObj.getValue("id", 0)

        if container.exists("pdbx_audit_revision_history"):
            tObj = container.getObj("pdbx_audit_revision_history")
            # Assuming the default sorting order from the release module -
            releaseDate = tObj.getValue("revision_date", 0)
        #
        if container.exists("chem_comp"):
            cObj = container.getObj("chem_comp")
            for ii in range(cObj.getRowCount()):
                ccId = cObj.getValue("id", ii)
                ccS.add(ccId)
        ret = ChemSummary(entryId, releaseDate, sorted(ccS))
        return ret

    def __scanContainer(self, container):
        """Scan the input container counting attributes and character width ranges and float precision."""
        cName = container.getName()
        loadStatusObj = container.getObj("rcsb_load_status")
        lName = loadStatusObj.getValue(attributeName="name", rowIndex=0)
        lFilePath = loadStatusObj.getValue(attributeName="locator", rowIndex=0)
        lDate = loadStatusObj.getValue(attributeName="load_date", rowIndex=0)
        #
        oD = {}
        for objName in container.getObjNameList():
            if objName == "rcsb_load_status":
                continue
            obj = container.getObj(objName)
            afD = self.__attributeDataTypeD[objName] if objName in self.__attributeDataTypeD else {}
            atNameList = obj.getAttributeList()
            wMin = {atName: 100000 for atName in atNameList}
            wMax = {atName: -1 for atName in atNameList}
            pMin = {atName: 100000 for atName in atNameList}
            pMax = {atName: -1 for atName in atNameList}
            for row in obj.getRowList():
                for ii, val in enumerate(row):
                    valLen = len(val)
                    if (valLen == 0) or (val == "?") or (val == "."):
                        continue
                    atName = atNameList[ii]
                    wMin[atName] = min(wMin[atName], valLen)
                    wMax[atName] = max(wMax[atName], valLen)
                    if atName in afD and afD[atName] == "float":
                        vPrec = 0
                        try:
                            fields = val.split(".")
                            vPrec = len(fields[1])
                            pMin[atName] = min(pMin[atName], vPrec)
                            pMax[atName] = max(pMax[atName], vPrec)
                        except Exception as e:
                            logger.debug("Failed to process float %s %r %r %s", atName, val, vPrec, str(e))
                            pMin[atName] = 0
                            pMax[atName] = 0
                        logger.debug("Got float for %s %r %r", atName, val, vPrec)
                    else:
                        pMin[atName] = 0
                        pMax[atName] = 0

            # ScanValue - containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec
            oD[objName] = [ScanValue(cName, objName, atN, wMin[atN], wMax[atN], pMin[atN], pMax[atN]) for atN in wMax if wMax[atN] != -1]
        # ScanSummary containerId, fromPath, scanCategoryDict
        #
        ret = ScanSummary(lName, lFilePath, lDate, oD)
        #
        return ret

    def __begin(self, message=""):
        startTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting %s at %s", message, ts)
        return startTime

    def __end(self, startTime, message=""):
        endTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        delta = endTime - startTime
        logger.debug("Completed %s at %s (%.4f seconds)", message, ts, delta)



##
#  File:  UpdateHoldingsProvider.py
#  Date:  18-Sep-2021 jdw
#
#  Updates:
#
##
"""Provide inventory of current repository update.
"""

import collections
import logging
import os.path

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class UpdateHoldingsProvider(object):
    """Provide inventory of the current repository update."""

    def __init__(self, cachePath, useCache, **kwargs):
        self.__dirPath = os.path.join(cachePath, "holdings")
        #
        baseUrl = kwargs.get("updateTargetUrl", "https://ftp.wwpdb.org/pub/pdb/data/status/latest")
        fallbackUrl = kwargs.get("updateFallbackUrl", "https://ftp.wwpdb.org/pub/pdb/data/status/latest")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__updD = self.__reloadUpdateLists(baseUrl, fallbackUrl, self.__dirPath, useCache=useCache)

    def testCache(self, minCount=100):
        logger.info("Length update updD (%d)", len(self.__updD) if self.__updD else 0)
        if self.__updD and len(self.__updD) > minCount:
            return True
        return False

    def getUpdateData(self):
        return self.__updD

    def __reloadUpdateLists(self, urlTarget, urlFallbackTarget, dirPath, useCache=True):
        """Parse legacy lists defining the contents of the repository update

        Args:
            urlTarget (str): base url for ftp repository instance
            urlFallbackTarget (str): fallback base url for ftp repository instance
            dirPath (str): cache directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing rcsb_repository_holdings_update

        """
        _ = urlFallbackTarget
        contentTypeList = ["pdb", "mr", "cs", "sf", "nmrdata"]
        contentNameD = {
            "pdb": "coordinates",
            "mr": "NMR restraints",
            "cs": "NMR chemical shifts",
            "sf": "structure factors",
            "nmrdata": "Combined NMR data (NEF)"
            # "nmrdata": "Combined NMR data (NMR-STAR)",
        }
        retD = {}
        fp = os.path.join(dirPath, "update_holdings.json")
        if useCache and self.__mU.exists(fp):
            retD = self.__mU.doImport(fp, fmt="json")
            logger.debug("Reading update cached IDs  (%d)", len(retD))
        else:
            try:
                updateTypeList = ["added", "modified", "obsolete"]
                #
                for updateType in updateTypeList:
                    uD = {}
                    for contentType in contentTypeList:
                        fp1 = os.path.join(urlTarget, updateType + "." + contentType)
                        if not self.__mU.exists(fp1):
                            logger.debug("skipping missing resource %r", fp1)
                            continue
                        entryIdL = self.__mU.doImport(fp1, "list")
                        #
                        for entryId in entryIdL:
                            entryId = entryId.strip().upper()
                            uD.setdefault(entryId, []).append(contentNameD[contentType])
                            if contentType == "nmrdata":
                                uD.setdefault(entryId, []).append("Combined NMR data (NMR-STAR)")
                    for entryId in uD:
                        uType = "removed" if updateType == "obsolete" else updateType
                        if entryId in retD:
                            logger.warning("Redundant entry in update list (%r) %r", updateType, entryId)
                        retD[entryId] = {"update_type": uType, "repository_content_types": uD[entryId]}
                ok = self.__mU.doExport(fp, retD, fmt="json", indent=3)
                logger.info("Stored update holdings (%r) for %d entries in %r", ok, len(retD), fp)

                tD = collections.Counter([v["update_type"] for v in retD.values()])
                logger.info("Update counts %r", tD)

            except Exception as e:
                logger.exception("Failing with %s", str(e))
        return retD
